@@ -49,6 +49,11 @@ from hypothesis.internal.reflection import arg_string, impersonate, \
 from hypothesis.internal.examplesource import ParameterSource
 
 
+def new_random():
+    import random
+    return random.Random(random.getrandbits(128))
+
+
 def time_to_call_it_a_day(settings, start_time):
     """Have we exceeded our timeout?"""
     if settings.timeout <= 0:
@@ -194,8 +199,6 @@ def simplify_template_such_that(
     it returned False.
 
     """
-    assert isinstance(random, Random)
-
     yield t
 
     if settings.max_shrinks <= 0 or not f(t):
@@ -410,11 +413,6 @@ def given(*generator_arguments, **generator_kwargs):
             return invalid(
                 u'given must be called with at least one argument')
 
-        if settings.derandomize:
-            random = Random(function_digest(test))
-        else:
-            random = provided_random or Random()
-
         if generator_arguments and original_argspec.varargs:
             return invalid(
                 u'varargs are not supported with positional arguments to '
@@ -478,6 +476,11 @@ def given(*generator_arguments, **generator_kwargs):
             test.__name__, argspec
         )
         def wrapped_test(*arguments, **kwargs):
+            if settings.derandomize:
+                random = Random(function_digest(test))
+            else:
+                random = provided_random or new_random()
+
             import hypothesis.strategies as sd
             from hypothesis.internal.strategymethod import strategy
 
@@ -515,9 +518,9 @@ def given(*generator_arguments, **generator_kwargs):
                 selfy = None
             test_runner = executor(selfy)
 
-            for example in getattr(
+            for example in reversed(getattr(
                 wrapped_test, u'hypothesis_explicit_examples', ()
-            ):
+            )):
                 if example.args:
                     example_kwargs = dict(zip(
                         argspec.args[-len(example.args):], example.args
@@ -535,11 +538,14 @@ def given(*generator_arguments, **generator_kwargs):
                     test.__name__, arg_string(test, arguments, example_kwargs)
                 )
                 try:
-                    test_runner(
-                        lambda: test(*arguments, **example_kwargs)
-                    )
+                    with BuildContext() as b:
+                        test_runner(
+                            lambda: test(*arguments, **example_kwargs)
+                        )
                 except BaseException:
                     report(message_on_failure)
+                    for n in b.notes:
+                        report(n)
                     raise
 
             if not any(
@@ -591,6 +597,7 @@ def given(*generator_arguments, **generator_kwargs):
 
             if perform_health_check:
                 initial_state = getglobalrandomstate()
+                health_check_random = Random(random.getrandbits(128))
                 with Settings(settings, verbosity=Verbosity.quiet):
                     count = 0
                     bad_draws = 0
@@ -603,8 +610,10 @@ def given(*generator_arguments, **generator_kwargs):
                             test_runner(reify_and_execute(
                                 search_strategy,
                                 search_strategy.draw_template(
-                                    random,
-                                    search_strategy.draw_parameter(random)),
+                                    health_check_random,
+                                    search_strategy.draw_parameter(
+                                        health_check_random,
+                                    )),
                                 lambda *args, **kwargs: None,
                             ))
                             count += 1
@@ -669,7 +678,9 @@ def given(*generator_arguments, **generator_kwargs):
                         'prevents Hypothesis from working correctly. '
                         'If you want to use methods from random, use '
                         'randoms() from hypothesis.strategies to get an '
-                        'instance of Random you can use.'
+                        'instance of Random you can use. Alternatively, you '
+                        'can use the random_module() strategy to explicitly '
+                        'seed the random module.'
                     )
 
             last_exception = [None]
@@ -711,7 +722,9 @@ def given(*generator_arguments, **generator_kwargs):
                             'Your test used the global random module. '
                             'This is unlikely to work correctly. You should '
                             'consider using the randoms() strategy from '
-                            'hypothesis.strategies instead.'
+                            'hypothesis.strategies instead. Alternatively, '
+                            'you can use the random_module() strategy to '
+                            'explicitly seed the random module.'
                         )
             is_template_example.__name__ = test.__name__
             is_template_example.__qualname__ = qualname(test)
@@ -768,7 +781,7 @@ def find(specifier, condition, settings=None, random=None, storage=None):
             )
         )
 
-    random = random or Random()
+    random = random or new_random()
     successful_examples = [0]
 
     def template_condition(template):
