@@ -20,22 +20,18 @@ to really unreasonable lengths to produce pretty output."""
 
 from __future__ import division, print_function, absolute_import
 
-import os
 import re
 import ast
-import sys
-import uuid
 import types
 import hashlib
 import inspect
+from types import ModuleType
 from functools import wraps
-from contextlib import contextmanager
 
 from hypothesis.settings import storage_directory
 from hypothesis.internal.compat import hrange, qualname, text_type, \
     getargspec, to_unicode, isidentifier, unicode_safe_repr, \
-    ARG_NAME_ATTRIBUTE, update_code_location, \
-    importlib_invalidate_caches
+    ARG_NAME_ATTRIBUTE, update_code_location
 
 
 def fully_qualified_name(f):
@@ -321,7 +317,7 @@ def arg_string(f, args, kwargs, reorder=True):
             bits.append(u'%s=%s' % (a, nicerepr(kwargs[a])))
 
     return u', '.join(
-        [unicode_safe_repr(x) for x in args] +
+        [nicerepr(x) for x in args] +
         bits
     )
 
@@ -342,13 +338,6 @@ def eval_directory():
     return storage_directory('eval_source')
 
 
-@contextmanager
-def add_directory_to_path(d):
-    sys.path.insert(0, d)
-    yield
-    sys.path.remove(d)
-
-
 eval_cache = {}
 
 
@@ -358,42 +347,12 @@ def source_exec_as_module(source):
     except KeyError:
         pass
 
-    d = eval_directory()
-    with add_directory_to_path(d):
-        final_name = u'hypothesis_temporary_module_%s' % (
-            hashlib.sha1(source.encode(u'utf-8')).hexdigest(),
-        )
-        temporary_name = u'hypothesis_temporary_module_%s_%s' % (
-            hashlib.sha1(source.encode(u'utf-8')).hexdigest(),
-            uuid.uuid4(),
-        )
-        temporary_filepath = os.path.join(d, temporary_name + u'.py')
-        final_filepath = os.path.join(d, final_name + u'.py')
-        f = open(temporary_filepath, u'w')
-        f.write(source)
-        f.close()
-        assert os.path.exists(temporary_filepath)
-
-        try:
-            os.rename(temporary_filepath, final_filepath)
-        except OSError:  # pragma: no cover
-            # The odds of final_filepath being a directory are basically zero,
-            # and it's basically impossible for them to be on different
-            # filesystems, so
-            # if this is raised it's because the destination already exists on
-            # Windows. That's fine, it won't be different, so just keep going,
-            # deleting our tempfile.
-            assert not os.path.isdir(final_filepath)
-            os.remove(temporary_filepath)
-
-        assert os.path.exists(final_filepath)
-        assert not os.path.exists(temporary_filepath)
-        with open(final_filepath) as r:
-            assert r.read() == source
-        importlib_invalidate_caches()
-        result = __import__(final_name)
-        eval_cache[source] = result
-        return result
+    result = ModuleType('hypothesis_temporary_module_%s' % (
+        hashlib.sha1(source.encode(u'utf-8')).hexdigest(),
+    ))
+    exec(source, result.__dict__)
+    eval_cache[source] = result
+    return result
 
 
 COPY_ARGSPEC_SCRIPT = """
