@@ -26,6 +26,7 @@ from hypothesis.settings import Settings
 from hypothesis.searchstrategy import SearchStrategy
 from hypothesis.internal.compat import ArgSpec, text_type, getargspec, \
     integer_types, float_to_decimal, unicode_safe_repr
+from hypothesis.internal.reflection import proxies
 from hypothesis.searchstrategy.reprwrapper import ReprWrapperStrategy
 
 __all__ = [
@@ -39,13 +40,61 @@ __all__ = [
     'dictionaries', 'fixed_dictionaries',
     'sampled_from',
     'builds',
-    'randoms',
+    'randoms', 'random_module',
     'streaming', 'basic', 'recursive', 'composite',
 ]
 
 
+class FloatKey(object):
+
+    def __init__(self, f):
+        self.value = float_to_int(f)
+
+    def __eq__(self, other):
+        return isinstance(other, FloatKey) and (
+            other.value == self.value
+        )
+
+    def __ne__(self, other):
+        return not self.__eq__(other)
+
+    def __hash__(self):
+        return hash(self.value)
+
+
+def convert_value(v):
+    if isinstance(v, float):
+        return FloatKey(v)
+    return (type(v), v)
+
+
+def cacheable(fn):
+    import weakref
+
+    cache = weakref.WeakValueDictionary()
+
+    @proxies(fn)
+    def cached_strategy(*args, **kwargs):
+        kwargs_cache_key = set()
+        try:
+            for k, v in kwargs.items():
+                kwargs_cache_key.add((k, convert_value(v)))
+        except TypeError:
+            return fn(*args, **kwargs)
+        cache_key = (
+            tuple(map(convert_value, args)), frozenset(kwargs_cache_key))
+        try:
+            return cache[cache_key]
+        except TypeError:
+            return fn(*args, **kwargs)
+        except KeyError:
+            result = fn(*args, **kwargs)
+            cache[cache_key] = result
+            return result
+    return cached_strategy
+
+
 def defines_strategy(strategy_definition):
-    from hypothesis.internal.reflection import proxies
     from hypothesis.searchstrategy.deferred import DeferredStrategy
 
     @proxies(strategy_definition)
@@ -88,6 +137,7 @@ def one_of(arg, *args):
     return OneOfStrategy(args)
 
 
+@cacheable
 @defines_strategy
 def integers(min_value=None, max_value=None):
     """Returns a strategy which generates integers (in Python 2 these may be
@@ -125,6 +175,7 @@ def integers(min_value=None, max_value=None):
             return BoundedIntStrategy(min_value, max_value)
 
 
+@cacheable
 @defines_strategy
 def booleans():
     """Returns a strategy which generates instances of bool."""
@@ -160,16 +211,34 @@ def int_to_float(value):
     )
 
 
+@cacheable
 @defines_strategy
-def floats(min_value=None, max_value=None):
-    """Returns a strategy which generates floats. If min_value is not None,
-    all values will be >= min_value. If max_value is not None, all values will
-    be <= max_value.
+def floats(
+    min_value=None, max_value=None, allow_nan=None, allow_infinity=None
+):
+    """Returns a strategy which generates floats.
+
+    - If min_value is not None, all values will be >= min_value.
+    - If max_value is not None, all values will be <= max_value.
+    - If min_value or max_value is not None, it is an error to enable
+      allow_nan.
+    - If both min_value and max_value are not None, it is an error to enable
+      allow_infinity.
 
     Where not explicitly ruled out by the bounds, all of infinity, -infinity
     and NaN are possible values generated by this strategy.
+
     """
 
+    if allow_nan is None:
+        allow_nan = bool(min_value is None and max_value is None)
+    elif allow_nan:
+        if min_value is not None or max_value is not None:
+            raise InvalidArgument(
+                u'Cannot have allow_nan=%r, with min_value or max_value' % (
+                    allow_nan
+                ))
+
     for e in (min_value, max_value):
         if e is not None and math.isnan(e):
             raise InvalidArgument(u'nan is not a valid end point')
@@ -182,6 +251,16 @@ def floats(min_value=None, max_value=None):
     if max_value == float(u'inf'):
         max_value = None
 
+    if allow_infinity is None:
+        allow_infinity = bool(min_value is None or max_value is None)
+    elif allow_infinity:
+        if min_value is not None and max_value is not None:
+            raise InvalidArgument(
+                u'Cannot have allow_infinity=%r, with both min_value and '
+                u'max_value' % (
+                    allow_infinity
+                ))
+
     from hypothesis.searchstrategy.numbers import WrapperFloatStrategy, \
         GaussianFloatStrategy, BoundedFloatStrategy, ExponentialFloatStrategy,\
         JustIntFloats, NastyFloats, FullRangeFloats, \
@@ -192,8 +271,8 @@ def floats(min_value=None, max_value=None):
             BoundedFloatStrategy() |
             ExponentialFloatStrategy() |
             JustIntFloats() |
-            NastyFloats() |
-            FullRangeFloats()
+            NastyFloats(allow_nan, allow_infinity) |
+            FullRangeFloats(allow_nan, allow_infinity)
         )
     elif min_value is not None and max_value is not None:
         if max_value < min_value:
@@ -239,30 +318,35 @@ def floats(min_value=None, max_value=None):
                 int_to_float
             )
     elif min_value is not None:
-        critical_values = [min_value, float(u'inf')]
+        critical_values = [min_value]
+        if allow_infinity:
+            critical_values.append(float(u'inf'))
         if is_negative(min_value):
             critical_values.append(-0.0)
         if min_value <= 0:
             critical_values.append(0.0)
         return (
-            floats().map(
+            floats(allow_infinity=allow_infinity, allow_nan=False).map(
                 lambda x: assume(not math.isnan(x)) and min_value + abs(x)
             )
         ) | sampled_from(critical_values)
     else:
         assert max_value is not None
-        critical_values = [max_value, float(u'-inf')]
+        critical_values = [max_value]
+        if allow_infinity:
+            critical_values.append(float(u'-inf'))
         if max_value >= 0:
             critical_values.append(-0.0)
             if not is_negative(max_value):
                 critical_values.append(0.0)
         return (
-            floats().map(
+            floats(allow_infinity=allow_infinity, allow_nan=False).map(
                 lambda x: assume(not math.isnan(x)) and max_value - abs(x)
             )
         ) | sampled_from(critical_values)
 
 
+@cacheable
 @defines_strategy
 def complex_numbers():
     """Returns a strategy that generates complex numbers."""
@@ -272,6 +356,7 @@ def complex_numbers():
     )
 
 
+@cacheable
 @defines_strategy
 def tuples(*args):
     """Return a strategy which generates a tuple of the same length as args by
@@ -310,6 +395,7 @@ def sampled_from(elements):
         return SampledFromStrategy(elements)
 
 
+@cacheable
 @defines_strategy
 def lists(
     elements=None, min_size=None, average_size=None, max_size=None,
@@ -411,6 +497,7 @@ def lists(
         )
 
 
+@cacheable
 @defines_strategy
 def sets(elements=None, min_size=None, average_size=None, max_size=None):
     """This has the same behaviour as lists, but returns sets instead.
@@ -426,6 +513,7 @@ def sets(elements=None, min_size=None, average_size=None, max_size=None):
     ).map(set)
 
 
+@cacheable
 @defines_strategy
 def frozensets(elements=None, min_size=None, average_size=None, max_size=None):
     """This is identical to the sets function but instead returns
@@ -454,6 +542,7 @@ def fixed_dictionaries(mapping):
     return FixedKeysDictStrategy(mapping)
 
 
+@cacheable
 @defines_strategy
 def dictionaries(
     keys, values, dict_class=dict,
@@ -490,6 +579,7 @@ def dictionaries(
     ).map(dict_class)
 
 
+@cacheable
 @defines_strategy
 def streaming(elements):
     """Generates an infinite stream of values where each value is drawn from
@@ -504,6 +594,7 @@ def streaming(elements):
     return StreamStrategy(elements)
 
 
+@cacheable
 @defines_strategy
 def characters(whitelist_categories=None, blacklist_categories=None,
                blacklist_characters=None, min_codepoint=None,
@@ -530,6 +621,7 @@ def characters(whitelist_categories=None, blacklist_categories=None,
                                  max_codepoint=max_codepoint)
 
 
+@cacheable
 @defines_strategy
 def text(
     alphabet=None,
@@ -566,6 +658,7 @@ def text(
     ))
 
 
+@cacheable
 @defines_strategy
 def binary(
     min_size=None, average_size=None, max_size=None
@@ -612,6 +705,7 @@ def basic(
     )
 
 
+@cacheable
 @defines_strategy
 def randoms():
     """Generates instances of Random (actually a Hypothesis specific
@@ -620,6 +714,45 @@ def randoms():
     return RandomStrategy(integers())
 
 
+class RandomSeeder(object):
+
+    def __init__(self, seed):
+        self.seed = seed
+
+    def __repr__(self):
+        return 'random.seed(%r)' % (self.seed,)
+
+
+@cacheable
+@defines_strategy
+def random_module():
+    """If your code depends on the global random module then you need to use
+    this.
+
+    It will explicitly seed the random module at the start of your test
+    so that tests are reproducible. The value it passes you is an opaque
+    object whose only useful feature is that its repr displays the
+    random seed. It is not itself a random number generator. If you want
+    a random number generator you should use the randoms() strategy
+    which will give you one.
+
+    """
+    from hypothesis.control import cleanup
+    import random
+
+    def seed_random(seed):
+        state = random.getstate()
+        random.seed(seed)
+        cleanup(lambda: random.setstate(state))
+        return RandomSeeder(seed)
+
+    return shared(
+        integers().map(seed_random),
+        'hypothesis.strategies.random_module()',
+    )
+
+
+@cacheable
 @defines_strategy
 def fractions():
     """Generates instances of fractions.Fraction."""
@@ -629,6 +762,7 @@ def fractions():
     )
 
 
+@cacheable
 @defines_strategy
 def decimals():
     """Generates instances of decimals.Decimal."""
@@ -640,6 +774,8 @@ def decimals():
     )
 
 
+@cacheable
+@defines_strategy
 def builds(target, *args, **kwargs):
     """Generates values by drawing from args and kwargs and passing them to
     target in the appropriate argument position.
@@ -649,26 +785,10 @@ def builds(target, *args, **kwargs):
     call target(i, flag=b).
 
     """
-    from hypothesis.internal.reflection import nicerepr
-
-    def splat(value):
-        return target(*value[0], **value[1])
-    target_name = getattr(target, u'__name__', type(target).__name__)
-    splat.__name__ = str(
-        u'splat(%s)' % (target_name,)
+    return tuples(tuples(*args), fixed_dictionaries(kwargs)).map(
+        lambda value: target(*value[0], **value[1])
     )
 
-    def calc_repr():
-        return u'builds(%s)' % (
-            u', '.join(
-                [nicerepr(target)] +
-                list(map(nicerepr, args)) +
-                sorted([u'%s=%r' % (k, v) for k, v in kwargs.items()])))
-
-    return ReprWrapperStrategy(
-        tuples(tuples(*args), fixed_dictionaries(kwargs)).map(splat),
-        calc_repr)
-
 
 @defines_strategy
 def recursive(base, extend, max_leaves=100):
@@ -719,6 +839,7 @@ def permutations(values):
     return lists(tuples(index, index), max_size=n ** 2).map(build_permutation)
 
 
+@cacheable
 def composite(f):
     """Defines a strategy that is built out of potentially arbitrarily many
     other strategies.
@@ -768,7 +889,7 @@ def composite(f):
 
 def shared(base, key=None):
     """Returns a strategy that draws a single shared value per run, drawn from
-    base. Any two shared instances with the samme key will share the same
+    base. Any two shared instances with the same key will share the same
     value, otherwise the identity of this strategy will be used. That is:
 
     >>> x = shared(s)
@@ -784,6 +905,7 @@ def shared(base, key=None):
     return SharedStrategy(base, key)
 
 
+@cacheable
 def choices():
     """Strategy that generates a function that behaves like random.choice.
 
@@ -825,6 +947,7 @@ def choices():
         ), 'chooser()')
 
 
+@cacheable
 def uuids():
     """Returns a strategy that generates UUIDs.
 
